local USE_DIFF = USE_DIFF or not DONT_USE_DIFF

if USE_DIFF then
	dofile("diff.lua")
end

local function getline(f)
	local s = f:read()
	if not s then return s end

	while s:sub(-1, -1) == '\r' or s:sub(-1, -1) == '\n' do
		s = s:sub(1, -2)
	end
	s = s:gsub("//.*", "")
	s = s:gsub("^%s*", "")
	
	-- s = s:gsub("^var ", "")
	
	local c = s:match("^var className%s*=%s*(\".*\");")
	if c then
		s = string.format("className = %s\n", c)
		return s
	end
		
	local name, v = s:match("(.*)%[i%] = (.*); i%+%+;?")
	
	if name then
		s = string.format("insert(%s, %s)", name, v:gsub("%[", "{"):gsub("]", "}"))
	elseif s:match("^%s*%]%s*;") then
		s = "})"
	elseif 
			s:match("[it] = 0;") or 
			s:match("^var%s+.*") or
			s:match("[it]%s*%+%+;") or
			s:match("^treeStartStop")
			then
		
		s = ""
	else
		name = s:match("(.*)%[i%]%s*=%s*%[")
		if name then
			s = string.format("insert(%s, {", name)
		else
			local value = s:match("\"([^\"]*)\"%s*];?")
			if value then
				s = string.format("%q})", value)
			end
		end
		local value
	end
	-- s = s:gsub("= %[([^]]+)%]", "= {%1}")
	
	-- print(s)
	
	return s .."\n"
end

local function loadjsdata(src, tbl)
	f = io.open(src, "r")
	
	local line
	local result, error = load(function () line = getline(f) return line end, src)
	if not result then
		print(string.format("%q\nLoading error : %s\n", line or src, error))
	else
		setfenv(result, tbl)
		local res, error = pcall(result)
		if not res then
			print(string.format("Execution error : %s\n", error))
		end
	end
end



local function conv_data(class, lang)
	local src = string.format("data-%s-%s.js", class, lang)
	local t = {
		tree = {},
		talent = {},
		rank = {},
		insert = table.insert,
		i = 0,
	}

	t.getTalentID = function (s)
		for i = 1, #t.talent do
			if t.talent[i][2] == s then
				return i
			end
		end
		error(string.format("unable to find talent \"%s\"", s))
	end
	
	loadjsdata(src, t)
	
	t.getTalentID = nil
	t.insert = nil
	t.class = class
	t.lang = lang
	
	return t
end

local function cleanup_html(src)
	local result = src:gsub("\r", ""):gsub("\n", ""):gsub("&nbsp;?", " ")
	local e = result:len()
	while e >= 1 and (" \t"):find(result:sub(e, e), 1, 1) do
		e = e - 1
	end
	return result:sub(1, e)
end

local function conv_tip(tip)
	if not tip:find("<", 1, 1) then
		return cleanup_html(tip)
	end
	local r = {}
	local p = 1
	local pspan = false
	local m = tip:len() + 1
	while p < m do
		local s1, e1, style, v1 = tip:find("<span style=([^>]+)>([^<]+)</span>", p)
		local s2, e2, v2 = tip:find("<span>([^<]+)</span>", p)
		local s3, e3 = tip:find("<[bB][rR]/?>", p)
		if not s1 and not s2 and not s3 then break end
		s1 = s1 or m
		s2 = s2 or m
		s3 = s3 or m
		
		if s1 < s2 and s1 < s3 then
			local t = style:match("text%-align:([^;]+)")
			if t and t == "right" then
				r[#r].right = cleanup_html(v1)
			else
				local s = cleanup_html(v1)
				local c = style:match("color:([^;]+)")
				if c then
					s = "|cff"..c..s.."|r"
				end
				r[#r + 1] = { left = s }
			end
			p = e1 + 1
			pspan = true
		elseif s2 < s3 then
			r[#r + 1] = { left = cleanup_html(v2) }
			pspan = true
			p = e2 + 1
		else
			v = tip:sub(p, s3 - 1)
			if v:len() > 0 or not pspan then
				r[#r + 1] = { left = cleanup_html(v) }
			end
			pspan = false
			p = e3 + 1
		end
	end
	local v = cleanup_html(tip:sub(p, -1))
	
	if v:len() then
		r[#r + 1] = { left = v }
	end
	
	for i = #r, 1, -1 do
		if not r[i].right and r[i].left == "" then
			r[i] = nil
		else
			break
		end
	end
	--[[
	if r[1].left == "30 yd range" then
		print("SRC", tip)
		for i, s in ipairs(r) do
			print("DST", i, s.left, s.right or s.color or "")
		end
	end
	--]]
	return r
end


local function check_tip_ending(tips)
	local has_terminal_point
	for _, tip in ipairs(tips) do
		if type(tip) ~= "string" then return end
		if tip:sub(-1) == "." then
			has_terminal_point = true
		end
	end
	if has_terminal_point then
		for i, tip in ipairs(tips) do
			if tip:sub(-1) ~= "." then
				tips[i] = tip.."."
			end
		end
	end
	if USE_DIFF then
		local fmt, results = build_diffs(unpack(tips))
		if fmt then
			tips.format = fmt
			for i = 1, #results do
				tips[i] = results[i]
			end
		end
	end
end



local function write_data(src, f)
	local w = function (s, ...)
		if s:find("%", 1, 1) then
			s = string.format(s, ...)
		end
		f:write("\t"..s.."\n")
	end
	w("data = Talented_Data.%s", src.class:upper())
	w("-- data.name = %q", src.className)
	local tindex, tree
	for index, talent in ipairs(src.talent) do
		local t, name = talent[1] + 1, talent[2]
		if tree ~= t then
			w("tree = data[%d]", t)
			w("tree.info.name = %q", cleanup_html(src.tree[t]))
			tree = t
			tindex = 1
		end
		w("talent = tree.talents[%d]", tindex)
		tindex = tindex + 1
		w("talent.info.name = %q", cleanup_html(name))
		w("talent.info.tips = {")
		local tips = {}
		for t, tip in ipairs(src.rank[index]) do
			tips[t] = conv_tip(tip)
		end
		check_tip_ending(tips)
		if tips.format then
			w("\tformat = %q,", tips.format)
			for i, results in ipairs(tips) do
				local c = {}
				for _, v in ipairs(results) do
					if tostring(tonumber(v)) == v then
						c[#c + 1] = ("%s"):format(v)
					else	
						c[#c + 1] = ("%q"):format(v)
					end
				end
				if #c > 1 then
					c = "{"..table.concat(c, ", ").."}"
				else
					c = c[1]
				end
				if c ~= tostring(i) then
					w("\t[%d] = %s,", i, c)
				end
			end
		else
			for t, tip in ipairs(tips) do
				if type(tip) == "string" then
					w("\t[%d] = %q,", t, tip)
				else
					w("\t[%d] = {", t)
					for line, s in ipairs(tip) do
						if s.right then
							w("\t\t[%d] = { left = %q, right = %q },", line, s.left, s.right)
						else
							w("\t\t[%d] = { left = %q },", line, s.left)
						end
					end
					w("\t},")
				end
			end
		end
		w("}")
	end
end

local LANG_MAP = {
	en = "enGB",
	us = "enUS",
	fr = "frFR",
	de = "deDE",
	es = "esES",
	kr = "koKR",
}

local CLASSES = {
	warrior = true,
	warlock = true,
	druid = true,
	hunter = true,
	mage = true,
	priest = true,
	shaman = true,
	rogue = true,
	paladin = true,
}

local function do_data(lang)
	local lmap = LANG_MAP[lang]
	local f = io.open(string.format("../Talented_Data_%s.lua", lmap), "w")
	f:write(string.format([[
-- Generated Data : Do Not Modify
if Talented_Data and GetLocale() == %q then
	local data, tree, talent
]], lmap))
	local conv = {}
	for class in pairs(CLASSES) do
		conv[class] = conv_data(class, lang)
	end
	
	for class in pairs(CLASSES) do
		write_data(conv[class], f)
	end

	f:write("end\n")
end

for lang in pairs(LANG_MAP) do
	do_data(lang)
end
